/*
 * Spew oopses, WARNs, BUGs or panics from all CPUs at once
 *
 * Copyright 2015 Anton Blanchard, IBM Corporation <anton@au1.ibm.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/module.h>
#include <linux/kthread.h>

static char spew[32];
module_param_string(spew, spew, 32, S_IRUGO);
MODULE_PARM_DESC(spew, "What to spew: WARN, BUG, panic or oops (default warn)");

enum
{
	WARN,
	BUG,
	PANIC,
	OOPS,
} type;

static unsigned int num_cpus;
static atomic_t running;

static int warnstorm_thread(void *data)
{
	atomic_inc(&running);

	while (atomic_read(&running) < num_cpus) {
		if (need_resched())
			schedule();
	}

	switch (type) {
		case WARN:
			WARN(1, "warnstorm");
			break;
		case BUG:
			BUG();
			break;
		case PANIC:
			panic("warnstorm");
			break;
		case OOPS:
			*(unsigned long *)0 = 0;
			break;
	}

	return 0;
}

static int __init warnstorm_init(void)
{
	unsigned int cpu;

	num_cpus = num_online_cpus();

	type = WARN;

	if (!strcasecmp(spew, "bug"))
		type = BUG;
	else if (!strcasecmp(spew, "panic"))
		type = PANIC;
	else if (!strcasecmp(spew, "oops"))
		type = OOPS;

	for_each_online_cpu(cpu) {
		struct task_struct *p;
		p = kthread_create(warnstorm_thread, NULL,
				   "warnstorm/%u", cpu);
		if (IS_ERR(p)) {
			pr_err("kthread_create on CPU %d failed\n", cpu);
			atomic_inc(&running);
		} else {
			kthread_bind(p, cpu);
			wake_up_process(p);
		}
	}

	return 0;
}

static void __exit warnstorm_exit(void)
{
}

module_init(warnstorm_init)
module_exit(warnstorm_exit)

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Anton Blanchard");
MODULE_DESCRIPTION("Spew WARNs, BUGs, panics or oopses from all CPUs at once");
